package edu.umn.cs.kumarbio.shear;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Random;
import java.util.Scanner;
import java.util.concurrent.ExecutionException;

import net.sf.picard.sam.AddOrReplaceReadGroups;

import org.broadinstitute.sting.commandline.CommandLineProgram;
import org.broadinstitute.sting.gatk.CommandLineGATK;

import edu.umn.cs.kumarbio.Genome;

public class ShearSV {

  private String  prefix;
  private String  origBamFile;
  private String  origFaRefFile;
  private String  origBwaIndexPrefix;
  private String  origTwoBitRefFile;
  private String  crestDir;
  private boolean debug;
  private String  rgId;
  private String  rgLb;
  private String  rgPl;
  private String  rgPu;
  private String  rgSm;

  public ShearSV(String[] args) {
    debug = false;
    // Skip first arg
    for (int i = 1; i < args.length; i++) {
      if (args[i].equals("-p")) {
        prefix = args[++i];
      } else if (args[i].equals("-b")) {
        origBamFile = args[++i];
      } else if (args[i].equals("-f")) {
        origFaRefFile = args[++i];
      } else if (args[i].equals("-i")) {
        origBwaIndexPrefix = args[++i];
      } else if (args[i].equals("-t")) {
        origTwoBitRefFile = args[++i];
      } else if (args[i].equals("-d")) {
        debug = true;
      } else {
        Shear.error("Unknown argument: " + args[i]);
      }
    }
    crestDir = System.getenv("CREST_DIRECTORY");
    if (prefix == null) {
      Shear.error("Must provide a prefix with the '-p' argument");
    }
    if (origBamFile == null) {
      Shear.error("Must provide a BAM alignment file with the '-b' argument");
    }
    if (origFaRefFile == null) {
      Shear.error("Must provide a FASTA reference file with the '-f' argument");
    }
    if (origBwaIndexPrefix == null) {
      origBwaIndexPrefix = origFaRefFile;
    }
    if (origTwoBitRefFile == null) {
      Shear.error("Must provide a Two-Bit reference file with the '-t' argument");
    }
    if (crestDir == null) {
      Shear.error("Must specify the CREST directory via the CREST_DIRECTORY environment variable");
    }
    if (!(new File(origBamFile).exists())) {
      Shear.error("BAM alignment file does not exist: " + origBamFile);
    }
    if (!(new File(origFaRefFile).exists())) {
      Shear.error("FASTA reference file does not exist: " + origFaRefFile);
    }
    if (!(new File(origTwoBitRefFile).exists())) {
      Shear.error("Two-Bit reference file does not exist: " + origTwoBitRefFile);
    }
    if (!(new File(crestDir).exists())) {
      Shear.error("CREST directory does not exist: " + crestDir);
    }
    System.out.println("[SHEAR] Starting SHEAR-SV");
    System.out.println("[SHEAR] Run prefix:             " + prefix);
    System.out.println("[SHEAR] Alignment file:         " + origBamFile);
    System.out.println("[SHEAR] FASTA reference file:   " + origFaRefFile);
    System.out.println("[SHEAR] BWA index prefix:       " + origBwaIndexPrefix);
    System.out.println("[SHEAR] Two-Bit reference file: " + origTwoBitRefFile);
    System.out.println("[SHEAR] CREST directory:        " + crestDir);
    if (debug) {
      System.out.println("[SHEAR] Debugging mode:         ON");
    } else {
      System.out.println("[SHEAR] Debugging mode:         OFF");
    }
  }

  private ArrayList<CrestPrediction> getCrestPreds(String crestPredFile) throws FileNotFoundException {
    ArrayList<CrestPrediction> preds = new ArrayList<CrestPrediction>();
    Scanner sc = new Scanner(new File(crestPredFile));
    while (sc.hasNextLine()) {
      String predLine = sc.nextLine();
      String[] tokens = predLine.split("\t");
      // What about cross-chromosome events?
      String bp1region = tokens[0];
      long bp1 = Long.parseLong(tokens[1]);
      String bp1dir = tokens[2];
      String bp2region = tokens[4];
      long bp2 = Long.parseLong(tokens[5]);
      String bp2dir = tokens[6];
      String type = tokens[8];
      preds.add(new CrestPrediction(type, bp1region, bp1, bp1dir, 0, 0, bp2region, bp2, bp2dir, 0, 0, 0));
    }
    sc.close();
    return preds;
  }

  private ArrayList<CrestPrediction> filterCrestPreds(ArrayList<CrestPrediction> origPreds) {
    // SHOULD MAKE SURE SORTED
    // NEED TO CHECK FOR SVs WITH SAME LOC ON BOTH BREAKPOINTS
    ArrayList<CrestPrediction> newPreds = new ArrayList<CrestPrediction>();
    tryNextPred: for (int i = 0; i < origPreds.size(); i++) {
      CrestPrediction origPred = origPreds.get(i);
      for (int k = 0; k < newPreds.size(); k++) {
        CrestPrediction newPred = newPreds.get(k);
        if ((origPred.bp1 == newPred.bp1) && origPred.bp1region.equals(newPred.bp1region)) {
          if (Math.abs(origPred.bp2 - newPred.bp2) <= 10) {
            if (newPred.bp2sc < origPred.bp2sc) {
              newPreds.set(k, origPred);
            }
            continue tryNextPred;
          }
        } else if ((origPred.bp2 == newPred.bp2) && origPred.bp2region.equals(newPred.bp2region)) {
          if (Math.abs(origPred.bp1 - newPred.bp1) <= 10) {
            if (newPred.bp1sc < origPred.bp1sc) {
              newPreds.set(k, origPred);
            }
            continue tryNextPred;
          }
        }
      }
      newPreds.add(origPreds.get(i));
    }
    return newPreds;
  }

  private void outputResults(ArrayList<CrestPrediction> crestPreds, FileWriter sdiOut, FileWriter reportOut, Genome gen)
      throws IOException {
    for (int i = 0; i < crestPreds.size(); i++) {
      CrestPrediction pred = crestPreds.get(i);
      if (pred.type.equals("DEL")) {
        sdiOut.write(getSdiLine(pred.bp1region, pred.bp1 + 1,
            gen.getSequence(pred.bp1region, pred.bp1 + 1, pred.bp2 - 1), "-"));
        reportOut.write(String.format("DEL\t%s:%d\t%s:%d\t%.2f%%\n", pred.bp1region, pred.bp1, pred.bp2region,
            pred.bp2, pred.heterogeneityPercent));
        System.out.printf("[SHEAR]     DEL\t%s:%d\t%s:%d\t%.2f%%\n", pred.bp1region, pred.bp1, pred.bp2region,
            pred.bp2, pred.heterogeneityPercent);
      } else if (pred.type.equals("INS")) {
        sdiOut.write(getSdiLine(pred.bp1region, pred.bp2, "-", gen.getSequence(pred.bp1region, pred.bp2, pred.bp1)));
        reportOut.write(String.format("TDUP\t%s:%d\t%s:%d\t%.2f%%\n", pred.bp1region, pred.bp1, pred.bp2region,
            pred.bp2, pred.heterogeneityPercent));
        System.out.printf("[SHEAR]     TDUP\t%s:%d\t%s:%d\t%.2f%%\n", pred.bp1region, pred.bp1, pred.bp2region,
            pred.bp2, pred.heterogeneityPercent);
      } else if (pred.type.equals("ITX")) {
        // ADD SDI WRITE
        reportOut.write(String.format("ITX\t%s:%d\t%s:%d\t%.2f%%\n", pred.bp1region, pred.bp1, pred.bp2region,
            pred.bp2, pred.heterogeneityPercent));
        System.out.printf("[SHEAR]     ITX\t%s:%d\t%s:%d\t%.2f%%\n", pred.bp1region, pred.bp1, pred.bp2region,
            pred.bp2, pred.heterogeneityPercent);
        System.out
            .printf("[SHEAR]       Skipping ITX for SDI output (Just deletions and tandem duplications for now)\n");
      } else if (pred.type.equals("CTX")) {
        // ADD SDI WRITE
        reportOut.write(String.format("CTX\t%s:%d\t%s:%d\t%.2f%%\n", pred.bp1region, pred.bp1, pred.bp2region,
            pred.bp2, pred.heterogeneityPercent));
        System.out.printf("[SHEAR]     CTX\t%s:%d\t%s:%d\t%.2f%%\n", pred.bp1region, pred.bp1, pred.bp2region,
            pred.bp2, pred.heterogeneityPercent);
        System.out
            .printf("[SHEAR]       Skipping CTX for SDI output (Just deletions and tandem duplications for now)\n");
      } else {
        throw new IllegalArgumentException("The following SV type is not valid: " + pred.type);
      }
    }
  }

  private ArrayList<CrestPrediction> processCrestPreds(String bamFileName, String refFileName,
      ArrayList<CrestPrediction> crestPreds) throws Exception {
    for (int i = 0; i < crestPreds.size(); i++) {
      CrestPrediction pred = crestPreds.get(i);
      System.out.printf("[SHEAR] Processing: %s %s:%d(%s) - %s:%d(%s)\n", pred.type, pred.bp1region, pred.bp1,
          pred.bp1dir, pred.bp2region, pred.bp2, pred.bp2dir);
      if (pred.type.equals("DEL")) {
        crestPreds.set(i, processDel(bamFileName, refFileName, pred));
      } else if (pred.type.equals("INS")) {
        crestPreds.set(i, processIns(bamFileName, refFileName, pred));
      } else if (pred.type.equals("ITX")) {
        crestPreds.set(i, processItx(bamFileName, refFileName, pred));
      } else if (pred.type.equals("CTX")) {
        crestPreds.set(i, processCtx(bamFileName, refFileName, pred));
      } else if (pred.type.equals("INV")) {
        // INVERSION SKIPPED FOR NOW
      } else {
        throw new IllegalArgumentException("The following SV type is not valid: " + pred.type);
      }
    }
    return crestPreds;
  }

  private CrestPrediction processDel(String bamFileName, String refFileName, CrestPrediction pred) throws Exception {

    Random rand = new Random(System.currentTimeMillis());
    String tmpFile1 = "tmp-" + pred.bp1region + pred.bp1 + "-" + rand.nextInt(Integer.MAX_VALUE);
    String tmpFile2 = "tmp-" + pred.bp2region + pred.bp2 + "-" + rand.nextInt(Integer.MAX_VALUE);

    String[] gatkArgs1 = { "-T", "Heterogeneity", "-I", bamFileName, "-R", refFileName, "-clipLocus",
        String.valueOf(pred.bp1), "-clipDirection", "+", "-L", pred.bp1region + ":" + pred.bp1, "-o", tmpFile1, "-l",
        "ERROR" };
    if (debug) {
      gatkArgs1[gatkArgs1.length - 1] = "INFO";
    }
    CommandLineProgram.start(new CommandLineGATK(), gatkArgs1);

    String[] gatkArgs2 = { "-T", "Heterogeneity", "-I", bamFileName, "-R", refFileName, "-clipLocus",
        String.valueOf(pred.bp2), "-clipDirection", "-", "-L", pred.bp2region + ":" + pred.bp2, "-o", tmpFile2, "-l",
        "ERROR" };
    if (debug) {
      gatkArgs2[gatkArgs2.length - 1] = "INFO";
    }
    CommandLineProgram.start(new CommandLineGATK(), gatkArgs2);

    File bp1file = new File(tmpFile1);
    Scanner bp1scanner = new Scanner(bp1file);
    int bp1sc = bp1scanner.nextInt();
    int bp1span = bp1scanner.nextInt();
    bp1scanner.close();
    bp1file.delete();

    File bp2file = new File(tmpFile2);
    Scanner bp2scanner = new Scanner(bp2file);
    int bp2sc = bp2scanner.nextInt();
    int bp2span = bp2scanner.nextInt();
    bp2scanner.close();
    bp2file.delete();

    double per = (100 * ((double) bp1sc + (double) bp2sc))
        / ((double) bp1sc + (double) bp2sc + (((double) bp1span + (double) bp2span) / 2));
    if (per > 100) {
      per = 100;
    }
    pred.bp1sc = bp1sc;
    pred.bp1span = bp1span;
    pred.bp2sc = bp2sc;
    pred.bp2span = bp2span;
    pred.heterogeneityPercent = per;

    System.out.println(String.format("[SHEAR]     Estimated heterogeneity:                %.2f%%",
        pred.heterogeneityPercent));

    return pred;

  }

  private CrestPrediction processIns(String bamFileName, String refFileName, CrestPrediction pred) throws Exception {

    Random rand = new Random(System.currentTimeMillis());
    String tmpFile1 = "tmp-" + pred.bp1region + pred.bp1 + "-" + rand.nextInt(Integer.MAX_VALUE);
    String tmpFile2 = "tmp-" + pred.bp2region + pred.bp2 + "-" + rand.nextInt(Integer.MAX_VALUE);

    String[] gatkArgs1 = { "-T", "Heterogeneity", "-I", bamFileName, "-R", refFileName, "-clipLocus",
        String.valueOf(pred.bp2), "-clipDirection", "-", "-L", pred.bp2region + ":" + pred.bp2, "-o", tmpFile1, "-l",
        "ERROR" };
    if (debug) {
      gatkArgs1[gatkArgs1.length - 1] = "INFO";
    }
    CommandLineProgram.start(new CommandLineGATK(), gatkArgs1);

    String[] gatkArgs2 = { "-T", "Heterogeneity", "-I", bamFileName, "-R", refFileName, "-clipLocus",
        String.valueOf(pred.bp1), "-clipDirection", "+", "-L", pred.bp1region + ":" + pred.bp1, "-o", tmpFile2, "-l",
        "ERROR" };
    if (debug) {
      gatkArgs2[gatkArgs2.length - 1] = "INFO";
    }
    CommandLineProgram.start(new CommandLineGATK(), gatkArgs2);

    File bp1file = new File(tmpFile1);
    Scanner bp1scanner = new Scanner(bp1file);
    int bp1sc = bp1scanner.nextInt();
    int bp1span = bp1scanner.nextInt();
    bp1scanner.close();
    bp1file.delete();

    File bp2file = new File(tmpFile2);
    Scanner bp2scanner = new Scanner(bp2file);
    int bp2sc = bp2scanner.nextInt();
    int bp2span = bp2scanner.nextInt();
    bp2scanner.close();
    bp2file.delete();

    double per = (100 * ((double) bp1sc + (double) bp2sc)) / (((double) bp1span + (double) bp2span) / 2);
    if (per > 100) {
      per = 100;
    }
    pred.bp1sc = bp1sc;
    pred.bp1span = bp1span;
    pred.bp2sc = bp2sc;
    pred.bp2span = bp2span;
    pred.heterogeneityPercent = per;

    System.out.println(String.format("[SHEAR]     Estimated heterogeneity:                %.2f%%",
        pred.heterogeneityPercent));

    return pred;

  }

  private CrestPrediction processItx(String bamFileName, String refFileName, CrestPrediction pred) throws Exception {

    // FIX clip directions

    Random rand = new Random(System.currentTimeMillis());
    String tmpFile1 = "tmp-" + pred.bp1region + pred.bp1 + "-" + rand.nextInt(Integer.MAX_VALUE);
    String tmpFile2 = "tmp-" + pred.bp2region + pred.bp2 + "-" + rand.nextInt(Integer.MAX_VALUE);

    String[] gatkArgs1 = { "-T", "Heterogeneity", "-I", bamFileName, "-R", refFileName, "-clipLocus",
        String.valueOf(pred.bp1), "-clipDirection", "+", "-L", pred.bp1region + ":" + pred.bp1, "-o", tmpFile1, "-l",
        "ERROR" };
    if (debug) {
      gatkArgs1[gatkArgs1.length - 1] = "INFO";
    }
    CommandLineProgram.start(new CommandLineGATK(), gatkArgs1);

    String[] gatkArgs2 = { "-T", "Heterogeneity", "-I", bamFileName, "-R", refFileName, "-clipLocus",
        String.valueOf(pred.bp2), "-clipDirection", "-", "-L", pred.bp2region + ":" + pred.bp2, "-o", tmpFile2, "-l",
        "ERROR" };
    if (debug) {
      gatkArgs2[gatkArgs2.length - 1] = "INFO";
    }
    CommandLineProgram.start(new CommandLineGATK(), gatkArgs2);

    File bp1file = new File(tmpFile1);
    Scanner bp1scanner = new Scanner(bp1file);
    int bp1sc = bp1scanner.nextInt();
    int bp1span = bp1scanner.nextInt();
    bp1scanner.close();
    bp1file.delete();

    File bp2file = new File(tmpFile2);
    Scanner bp2scanner = new Scanner(bp2file);
    int bp2sc = bp2scanner.nextInt();
    int bp2span = bp2scanner.nextInt();
    bp2scanner.close();
    bp2file.delete();

    double per = (100 * ((double) bp1sc + (double) bp2sc)) / ((double) bp1sc + (double) bp2sc + bp1span + bp2span);
    if (per > 100) {
      per = 100;
    }
    pred.bp1sc = bp1sc;
    pred.bp1span = bp1span;
    pred.bp2sc = bp2sc;
    pred.bp2span = bp2span;
    pred.heterogeneityPercent = per;

    System.out.println(String.format("[SHEAR]     Estimated heterogeneity:                %.2f%%",
        pred.heterogeneityPercent));

    return pred;

  }

  private CrestPrediction processCtx(String bamFileName, String refFileName, CrestPrediction pred) throws Exception {

    // FIX clip directions

    Random rand = new Random(System.currentTimeMillis());
    String tmpFile1 = "tmp-" + pred.bp1region + pred.bp1 + "-" + rand.nextInt(Integer.MAX_VALUE);
    String tmpFile2 = "tmp-" + pred.bp2region + pred.bp2 + "-" + rand.nextInt(Integer.MAX_VALUE);

    String[] gatkArgs1 = { "-T", "Heterogeneity", "-I", bamFileName, "-R", refFileName, "-clipLocus",
        String.valueOf(pred.bp1), "-clipDirection", "+", "-L", pred.bp1region + ":" + pred.bp1, "-o", tmpFile1, "-l",
        "ERROR" };
    if (debug) {
      gatkArgs1[gatkArgs1.length - 1] = "INFO";
    }
    CommandLineProgram.start(new CommandLineGATK(), gatkArgs1);

    String[] gatkArgs2 = { "-T", "Heterogeneity", "-I", bamFileName, "-R", refFileName, "-clipLocus",
        String.valueOf(pred.bp2), "-clipDirection", "-", "-L", pred.bp2region + ":" + pred.bp2, "-o", tmpFile2, "-l",
        "ERROR" };
    if (debug) {
      gatkArgs2[gatkArgs2.length - 1] = "INFO";
    }
    CommandLineProgram.start(new CommandLineGATK(), gatkArgs2);

    File bp1file = new File(tmpFile1);
    Scanner bp1scanner = new Scanner(bp1file);
    int bp1sc = bp1scanner.nextInt();
    int bp1span = bp1scanner.nextInt();
    bp1scanner.close();
    bp1file.delete();

    File bp2file = new File(tmpFile2);
    Scanner bp2scanner = new Scanner(bp2file);
    int bp2sc = bp2scanner.nextInt();
    int bp2span = bp2scanner.nextInt();
    bp2scanner.close();
    bp2file.delete();

    double per = (100 * ((double) bp1sc + (double) bp2sc)) / ((double) bp1sc + (double) bp2sc + bp1span + bp2span);
    if (per > 100) {
      per = 100;
    }
    pred.bp1sc = bp1sc;
    pred.bp1span = bp1span;
    pred.bp2sc = bp2sc;
    pred.bp2span = bp2span;
    pred.heterogeneityPercent = per;

    System.out.println(String.format("[SHEAR]     Estimated heterogeneity:                %.2f%%",
        pred.heterogeneityPercent));

    return pred;

  }

  /*
   * Returns the properly formatted SDI line for a given region name, start
   * location, reference string, and consensus string.
   */
  public static String getSdiLine(String regionName, long startLoc, String reference, String consensus) {
    StringBuffer line = new StringBuffer();
    line.append(regionName);
    line.append("\t");
    line.append(startLoc);
    line.append("\t");
    if (reference.equals("-")) {
      if (consensus.equals("-")) {
        line.append(0);
      } else {
        line.append(consensus.length());
      }
    } else if (consensus.equals("-")) {
      if (reference.equals("-")) {
        line.append(0);
      } else {
        line.append("-");
        line.append(reference.length());
      }
    } else {
      line.append(consensus.length() - reference.length());
    }
    line.append("\t");
    line.append(reference);
    line.append("\t");
    line.append(consensus);
    line.append("\t*\t*\n");
    return line.toString();
  }

  private void saveReadGroupInfo(String bamFile) throws InterruptedException, IOException, ExecutionException {
    String tmpOutput = prefix + ".tmp" + System.currentTimeMillis();
    FileOutputStream output = new FileOutputStream(tmpOutput + ".origHeaders");
    RunCommand.run(new ProcessBuilder("samtools", "view", "-H", bamFile), output, System.out, "samtools");
    output.close();
    Scanner sc = new Scanner(new File(tmpOutput + ".origHeaders"));
    while (sc.hasNextLine()) {
      String line = sc.nextLine();
      if (line.startsWith("@RG")) {
        String[] tokens = line.split("\t");
        for (int i = 1; i < tokens.length; i++) {
          if (tokens[i].startsWith("ID")) {
            rgId = tokens[i].substring(3);
          } else if (tokens[i].startsWith("LB")) {
            rgLb = tokens[i].substring(3);
          } else if (tokens[i].startsWith("PL")) {
            rgPl = tokens[i].substring(3);
          } else if (tokens[i].startsWith("PU")) {
            rgPu = tokens[i].substring(3);
          } else if (tokens[i].startsWith("SM")) {
            rgSm = tokens[i].substring(3);
          }
        }
        if (rgId == null) {
          rgId = prefix;
        }
        if (rgLb == null) {
          rgLb = rgId;
        }
        if (rgPl == null) {
          rgPl = rgId;
        }
        if (rgPu == null) {
          rgPu = rgId;
        }
        if (rgSm == null) {
          rgSm = rgId;
        }
        break;
      }
    }
    sc.close();
    new File(tmpOutput + ".origHeaders").delete();
  }

  private void runCrest(String bamFile, String crestPrefix, String faRefFile, GFServer gfServer)
      throws InterruptedException, IOException, ExecutionException {
    ProcessBuilder extractSClipProcess = new ProcessBuilder("perl", "-I" + crestDir, crestDir + "/extractSClip.pl",
        "-p", crestPrefix, "-i", bamFile, "--ref_genome", faRefFile);
    ProcessBuilder crestProcess = new ProcessBuilder("perl", "-I" + crestDir, crestDir + "/CREST.pl", "-max_rep_cover",
        "1000000", "-p", crestPrefix, "-f", crestPrefix + ".cover", "-d", bamFile, "--ref_genome", faRefFile, "-t",
        gfServer.getTwoBitFile(), "--blatserver", "localhost", "--blatport", "" + gfServer.getPortNum());
    if (debug) {
      RunCommand.run(extractSClipProcess, System.out, "CREST");
      RunCommand.run(crestProcess, System.out, "CREST");
    } else {
      RunCommand.run(extractSClipProcess);
      RunCommand.run(crestProcess);
      new File(crestPrefix + ".sclip.txt").delete();
      new File(crestPrefix + ".cover").delete();
    }
  }

  private ArrayList<String> getBreakpoints(String crestPredFile) throws FileNotFoundException {
    Scanner sc = new Scanner(new File(crestPredFile));
    ArrayList<String> breakpoints = new ArrayList<String>();
    while (sc.hasNextLine()) {
      String line = sc.nextLine();
      String[] tokens = line.split("\t");
      breakpoints.add(tokens[0] + ":" + tokens[1]);
      breakpoints.add(tokens[4] + ":" + tokens[5]);
    }
    sc.close();
    return breakpoints;
  }

  private void extractRelevantReads(ArrayList<String> breakpoints, String bamFile, String faRefFile, String fqOutputFile)
      throws InterruptedException, IOException, Exception {
    String tmpOutput = prefix + ".tmp" + System.currentTimeMillis();
    String[] gatkArgs = new String[12 + (2 * breakpoints.size())];
    int ind = 0;
    gatkArgs[ind++] = "-T";
    gatkArgs[ind++] = "PrintReads";
    gatkArgs[ind++] = "-I";
    gatkArgs[ind++] = bamFile;
    gatkArgs[ind++] = "-R";
    gatkArgs[ind++] = faRefFile;
    gatkArgs[ind++] = "--interval_padding";
    gatkArgs[ind++] = "25";
    for (int i = 0; i < breakpoints.size(); i++) {
      gatkArgs[ind++] = "--intervals";
      gatkArgs[ind++] = breakpoints.get(i);
    }
    gatkArgs[ind++] = "-o";
    gatkArgs[ind++] = tmpOutput + ".bam";
    gatkArgs[ind++] = "-l";
    gatkArgs[ind++] = "ERROR";
    if (debug) {
      gatkArgs[gatkArgs.length - 1] = "INFO";
    }
    CommandLineProgram.start(new CommandLineGATK(), gatkArgs);
    FileOutputStream output;
    output = new FileOutputStream(tmpOutput + ".sam");
    RunCommand.run(new ProcessBuilder("samtools", "view", "-h", tmpOutput + ".bam"), output, System.out, "samtools");
    RunCommand.run(new ProcessBuilder("samtools", "view", "-f", "4", bamFile), output, System.out, "samtools");
    output.close();
    Scanner sc = new Scanner(new File(tmpOutput + ".sam"));
    PrintWriter writer = new PrintWriter(fqOutputFile);
    while (sc.hasNextLine()) {
      String line = sc.nextLine();
      if (line.charAt(0) != '@') {
        String[] tokens = line.split("\t");
        if (BigInteger.valueOf(Long.valueOf(tokens[1]).byteValue()).testBit(6)) {
          writer.println("@" + tokens[0] + "-1");
        } else {
          writer.println("@" + tokens[0] + "-2");
        }
        if (BigInteger.valueOf(Long.valueOf(tokens[1]).byteValue()).testBit(4)) {
          writer.println(Genome.rc(tokens[9]));
        } else {
          writer.println(tokens[9]);
        }
        writer.println("+");
        if (BigInteger.valueOf(Long.valueOf(tokens[1]).byteValue()).testBit(4)) {
          writer.println(Genome.reverse(tokens[10]));
        } else {
          writer.println(tokens[10]);
        }
      }
    }
    sc.close();
    writer.close();
    new File(tmpOutput + ".bam").delete();
    new File(tmpOutput + ".bai").delete();
    new File(tmpOutput + ".sam").delete();
  }

  public void localAlign(String fqFile, String faRefFile, String bwaIndexPrefix, String bamOutputFile)
      throws FileNotFoundException, InterruptedException, IOException, ExecutionException {
    String tmpOutput = prefix + ".tmp" + System.currentTimeMillis();
    FileOutputStream output = new FileOutputStream(tmpOutput + "-1.sam");
    RunCommand.run(new ProcessBuilder("bwa", "bwasw", "-M", bwaIndexPrefix, fqFile), output, System.out, "BWA-SW");
    output.close();
    // Sorting must be done before calmd, or else calmd will be very slow
    String[] picardArgs = { "I=" + tmpOutput + "-1.sam", "O=" + tmpOutput + "-2.bam", "SO=coordinate", "RGID=" + rgId,
        "RGLB=" + rgLb, "RGPL=" + rgPl, "RGPU=" + rgPu, "RGSM=" + rgSm, "QUIET=TRUE", "VERBOSITY=ERROR" };
    if (debug) {
      picardArgs[picardArgs.length - 2] = "QUIET=FALSE";
      picardArgs[picardArgs.length - 1] = "VERBOSITY=INFO";
    }
    new AddOrReplaceReadGroups().instanceMain(picardArgs);
    output = new FileOutputStream(tmpOutput + "-3.bam");
    // calmd is getting stuck! I think the error stream is getting filled up
    // before it has a chance to be read from. Need concurrent threads?
    RunCommand.run(new ProcessBuilder("samtools", "calmd", "-b", tmpOutput + "-2.bam", faRefFile), output, System.out,
        "samtools");
    output.close();
    output = new FileOutputStream(bamOutputFile);
    RunCommand.run(new ProcessBuilder("samtools", "view", "-b", "-F", "256", tmpOutput + "-3.bam"), output, System.out,
        "samtools");
    output.close();
    RunCommand.run(new ProcessBuilder("samtools", "index", bamOutputFile), System.out, "samtools");
    new File(tmpOutput + "-1.sam").delete();
    new File(tmpOutput + "-2.bam").delete();
    new File(tmpOutput + "-3.bam").delete();
  }

  public void execute() throws InterruptedException, IOException, Exception {
    GFServer gfServer = new GFServer(prefix, origTwoBitRefFile);
    System.out.println("[SHEAR] Using port " + gfServer.getPortNum() + " for gfServer");
    System.out.println("[SHEAR] Starting gfServer...");
    gfServer.start();
    System.out.println("[SHEAR] Waiting for gfServer to be ready...");
    gfServer.waitForReady();
    System.out.println("[SHEAR] gfServer is ready!");
    System.out.println("[SHEAR] Extracting original read group information...");
    saveReadGroupInfo(origBamFile);
    System.out.println("[SHEAR] Making SV Predictions with CREST...");
    runCrest(origBamFile, prefix + ".0", origFaRefFile, gfServer);
    System.out.println("[SHEAR] Extracting Breakpoints...");
    ArrayList<String> breakpoints = getBreakpoints(prefix + ".0.predSV.txt");
    System.out.println("[SHEAR] Breakpoints:");
    boolean hasNewBreakpoints = true;
    if (breakpoints.size() > 0) {
      for (int i = 0; i < breakpoints.size(); i++) {
        System.out.println("[SHEAR]   " + breakpoints.get(i));
      }
    } else {
      System.out.println("[SHEAR]   None");
      hasNewBreakpoints = false;
    }
    int ctr = 0;
    while (hasNewBreakpoints) {
      ctr++;
      System.out.println("[SHEAR] Extracting Breakpoint and Unmapped Reads...");
      extractRelevantReads(breakpoints, origBamFile, origFaRefFile, prefix + "." + ctr + ".extracted.fq");
      System.out.println("[SHEAR] Realigning using BWA-SW local alignment...");
      localAlign(prefix + "." + ctr + ".extracted.fq", origFaRefFile, origBwaIndexPrefix, prefix + "." + ctr
          + ".extracted.aligned.bam");
      if (!debug) {
        new File(prefix + "." + ctr + ".extracted.fq").delete();
      }
      System.out.println("[SHEAR] Making SV Predictions with CREST...");
      runCrest(prefix + "." + ctr + ".extracted.aligned.bam", prefix + "." + ctr + "", origFaRefFile, gfServer);
      System.out.println("[SHEAR] Extracting Breakpoints...");
      ArrayList<String> newBreakpoints = getBreakpoints(prefix + "." + ctr + ".predSV.txt");
      hasNewBreakpoints = false;
      for (int i = 0; i < newBreakpoints.size(); i++) {
        if (!breakpoints.contains(newBreakpoints.get(i))) {
          System.out.println("[SHEAR] New breakpoint: " + newBreakpoints.get(i));
          breakpoints.add(newBreakpoints.get(i));
          hasNewBreakpoints = true;
        }
      }
    }
    System.out.println("[SHEAR] No new breakpoints detected");
    gfServer.close();
    if (!debug) {
      new File(prefix + ".gfServer.log").delete();
      new File(prefix + ".gfServer.out").delete();
    }

    System.out.println("[SHEAR] Estimating Heterogeneity...");
    ArrayList<CrestPrediction> crestPreds = getCrestPreds(prefix + "." + ctr + ".predSV.txt");
    crestPreds = processCrestPreds(prefix + "." + ctr + ".extracted.aligned.bam", origFaRefFile, crestPreds);
    if (!debug) {
      for (int i = 0; i <= ctr; i++) {
        // Check to make sure files actually exist?
        new File(prefix + "." + i + ".predSV.txt").delete();
        new File(prefix + "." + i + ".extracted.aligned.bam").delete();
        new File(prefix + "." + i + ".extracted.aligned.bam.bai").delete();
      }
    }

    System.out.println("[SHEAR] Filtering Results...");
    crestPreds = filterCrestPreds(crestPreds);

    System.out.println("[SHEAR] Outputing Results...");
    Genome gen = new Genome(origFaRefFile);
    FileWriter sdiOut = new FileWriter(prefix + ".sdi");
    FileWriter reportOut = new FileWriter(prefix + ".report");
    outputResults(crestPreds, sdiOut, reportOut, gen);
    gen.close();
    sdiOut.close();
    reportOut.close();

    System.out.println("[SHEAR] All Finished!");
  }

}
